import sys

# Define credentials below

# DockerHub
# The DockerHub repository that will hold Docker Images pushed by the BugSwarm
# Reproducer. For example `bugswarm/images`.
DOCKER_HUB_REPO = '#'
# The DockerHub repository that will hold Docker Images pushed by the
# CacheDependencies step in the BugSwarm Reproducer. Leave this blank to skip
# caching dependencies. For example `bugswarm/cached-images`.
DOCKER_HUB_CACHED_REPO = '#'
# This is the username credential being used by the DockerHub API to access the
# DOCKER_HUB_REPO. For example `bugswarm`.
DOCKER_HUB_USERNAME = '#'
# This is the password credential being used by the DockerHub API to access the
# DOCKER_HUB_REPO. You can also use an access token instead.
DOCKER_HUB_PASSWORD = '#'

# Docker Private Registry
DOCKER_REGISTRY_REPO = '#'
DOCKER_REGISTRY_USERNAME = '#'
DOCKER_REGISTRY_PASSWORD = '#'

# GitHub
# A GitHub Access Token to perform Git Operations over HTTPS via the Git API.
# (used for Mining projects)
# These GitHub tokens are hard-coded and can be used for token switching to
# minimize the time spent waiting for our GitHub quota to reset.
GITHUB_TOKENS = '#'

# Travis
# A Travis CI Access Token to send authenticated requests to Travis CI.
# (used for gathering builds)
# Travis CI Access Token for sending authenticated requests up to 2000/min
# Unauthenticated requests are up to 500/min
TRAVIS_TOKENS = '#'

# Database authentication tokens
# The token of a user's account used to access MongoDB.
# ('testDBPassword' if using Docker image of Mongo)
DATABASE_PIPELINE_TOKEN = '#'

# This hostname is used to integrate the API usage with your local database.
# It should be `<LOCAL-IPADDRESS>:5000`, for example `127.0.0.1:5000`.
COMMON_HOSTNAME = 'www.api.bugswarm.org'

# Check validity of credentials

# DockerHub check
if not DOCKER_HUB_REPO:
    print('[ERROR]: DOCKER_HUB_REPO has not been found. Please input your credentials under '
          'common/credentials.py and rerun the bugswarm/provision.sh script.')
    sys.exit(1)
if not DOCKER_HUB_CACHED_REPO:
    print('[WARNING]: DOCKER_HUB_CACHED_REPO has not been found. Skip caching dependencies in reproducing stage.')
if not DOCKER_HUB_USERNAME:
    print('[ERROR]: DOCKER_HUB_USERNAME has not been found. Please input your credentials under '
          'common/credentials.py and rerun the bugswarm/provision.sh script.')
    sys.exit(1)
if not DOCKER_HUB_PASSWORD:
    print('[ERROR]: DOCKER_HUB_PASSWORD has not been found. Please input your credentials under '
          'common/credentials.py and rerun the bugswarm/provision.sh script.')
    sys.exit(1)

# Docker Private Registry check
if not DOCKER_REGISTRY_REPO:
    print('[WARNING]: DOCKER_REGISTRY_REPO has not been found. Skip pushing to docker private registry '
          'in reproducing stage')
if not DOCKER_REGISTRY_USERNAME:
    print('[WARNING]: DOCKER_REGISTRY_USERNAME has not been found. Skip pushing to docker private registry '
          'in reproducing stage')
if not DOCKER_REGISTRY_PASSWORD:
    print('[WARNING]: DOCKER_REGISTRY_PASSWORD has not been found. Skip pushing to docker private registry '
          'in reproducing stage')

# GitHub check
if not GITHUB_TOKENS:
    print('[ERROR]: GITHUB_TOKENS has not been set. Please input your credentials under common/credentials.py and '
          'rerun the bugswarm/provision.sh script.')
    sys.exit(1)

# Database authentication tokens check
if not DATABASE_PIPELINE_TOKEN:
    print('[ERROR]: DATABASE_PIPELINE_TOKEN has not been found. Please input your credentials under '
          'common/credentials.py and rerun the bugswarm/provision.sh script.')
    sys.exit(1)



import subprocess

from typing import Union

from .shell_wrapper import ShellWrapper

_CommitSHA = str


def get_image_tag(repo: str, failed_job_id: Union[str, int]) -> str:
    """
    Construct the unique image tag identifying the pair with the given repository slug and failed job ID.

    :param repo: A repository slug.
    :param failed_job_id: A failed job ID for the pair for which the image tag will represent.
    :return: An image tag.
    """
    if not isinstance(repo, str):
        raise TypeError
    if not (isinstance(failed_job_id, str) or isinstance(failed_job_id, int)):
        raise TypeError
    if repo.count('/') != 1:
        raise ValueError('The repository slug should contain exactly one slash.')
    return '{}-{}'.format(repo.replace('/', '-'), failed_job_id)


def get_current_component_version_message(component_name: str) -> str:
    """
    Get a message that can be logged to indicate the version of the currently executing BugSwarm component.

    :param component_name: The name of the component. This name is not used to obtain the version. Instead, it is merely
                           used to compose the returned message.
    :return: A string that can be logged to indicate the version of the currently executing BugSwarm component.
    """
    if not isinstance(component_name, str):
        raise TypeError
    version = _get_current_component_version()[:7]
    return 'Using version {} of {}.'.format(version, component_name)


def _get_current_component_version() -> _CommitSHA:
    stdout, _, _ = ShellWrapper.run_commands('git rev-parse HEAD', stdout=subprocess.PIPE, shell=True)
    return stdout


def get_diff_stats(repo, sha_fail, sha_pass, gh_wrapper) -> (int, int, int):
    """
    Get the diff size between two commits by using Github API
    :param repo: repo slug e.g. bugswarm/common
    :param sha_fail: Github failed commit sha
    :param sha_pass: Github passed commit sha
    :param gh_wrapper: Github_wrapper
    :return: additions, deletions, changes
    """

    url = 'https://api.github.com/repos/{}/compare/{}...{}'.format(repo, sha_fail, sha_pass)

    # Try to make github api request
    response, response_json = gh_wrapper.get(url)

    if response is None or response_json is None:
        return -1, -1, -1

    # Initialize additions, etc to 0.
    additions = 0
    deletions = 0
    changes = 0

    # For every file changed, get the additions, etc
    for f in response_json['files']:
        additions += f['additions']
        deletions += f['deletions']
        changes += f['changes']

    return additions, deletions, changes



import subprocess

from typing import Optional
from typing import Tuple


class ShellWrapper(object):

    @staticmethod
    def run_commands(*commands: str, **kwargs) -> Tuple[Optional[str], Optional[str], int]:
        """
        Run a list of commands sequentially in the same shell environment and wait for the commands to complete.
        All keyword arguments are passed to the subprocess.run function.

        :param commands: Strings that represent commands to run.
        :param kwargs: Keyword arguments that are passed to the Popen constructor.
        :return: A 3-tuple of the subprocess' stdout stream, stderr stream, and return code. The streams in the tuple
                 can be None depending on the passed values of `stdout` and `stderr`.
        """
        command = ' ; '.join(commands)
        # Indirectly waits for a return code.
        process = subprocess.run(command, **kwargs)
        stdout = process.stdout
        stderr = process.stderr
        # Decode stdout and stderr to strings if needed.
        if isinstance(stdout, bytes):
            stdout = str(stdout, 'utf-8').strip()
        if isinstance(stderr, bytes):
            stderr = str(stderr, 'utf-8').strip()
        return stdout, stderr, process.returncode



import os
from concurrent.futures import ThreadPoolExecutor
from concurrent.futures import as_completed
from distutils.dir_util import copy_tree
from typing import Callable
from typing import List

from . import utils as procutils
from bugswarm.common import log


class ParallelArtifactRunner(object):
    """
    ParallelArtifactRunner is an abstract base class that facilitates parallel processing of BugSwarm artifacts.

    A concrete subclass must override the process_artifact method where it can perform some operation on each artifact.
    A concrete subclass may also override the following methods:
    - pre_run:   Called before processing any artifacts
    - post_run:  Called after processing all artifacts
    """
    def __init__(self, image_tags: List[str], workers: int = 1):
        """
        :param image_tags: A list of image tags representing BugSwarm artifacts.
        :param workers: The maximum number of worker threads to spawn, clamped to (0, length of `image_tags`].
                        Defaults to 1, in which case each image tag is processed sequentially.
        """
        if not image_tags:
            raise ValueError
        if workers <= 0:
            raise ValueError
        self._image_tags = image_tags
        self._num_workers = min(workers, len(self._image_tags))

    def run(self):
        """
        Start processing image tags.

        Overriding is forbidden.
        """
        self.pre_run()
        with ThreadPoolExecutor(max_workers=self._num_workers) as executor:
            future_to_image_tag = {executor.submit(self._thread_main, image_tag): image_tag
                                   for image_tag in self._image_tags}
        attempted = 0
        succeeded = 0
        errored = 0
        for future in as_completed(future_to_image_tag):
            attempted += 1
            try:
                data = future.result()
                if data:
                    succeeded += 1
                else:
                    errored += 1
            except Exception as e:
                log.error(e)
                errored += 1
        self.post_run()

    def _thread_main(self, image_tag: str):
        return self.process_artifact(image_tag)

    def process_artifact(self, image_tag: str):
        """
        Subclasses must override this method to process each item in the workload.

        Subclasses that do not inherit directly from ParallelArtifactRunner can call super as appropriate.
        :param image_tag: The image tag representing the artifact to process.
        :return: The eventual return value of the worker thread that processed `image_tag`.
        """
        raise NotImplementedError

    def pre_run(self):
        """
        Called before any items have been processed.

        Overriding is optional. Defaults to no-op. Subclasses should call super as appropriate.
        """
        pass

    def post_run(self):
        """
        Called after all items have been processed.

        Overriding is optional. Defaults to no-op. Subclasses can call super as appropriate.
        """
        pass


class CopyAndExecuteArtifactRunner(ParallelArtifactRunner):
    """
    Concrete subclass of ParallelArtifactRunner that facilitates a common runner workflow:
    - List, in a file, BugSwarm artifact image tags to process;
    - Copy files, that should be available to all processed artifacts, into the host-side sandbox;
    - For each processed artifact, execute commands inside the artifact container.
    """
    def __init__(self, image_tags_file: str, copy_dir: str, command: Callable[[str], str], workers: int = 1):
        """
        :param image_tags_file: Path to a file containing a newline-separated list of image tags.
        :param copy_dir: A directory to copy into the host-side sandbox before any artifacts are processed.
        :param command: A callable used to determine what command(s) to execute in each artifact container. `command` is
                        called once for each processed artifact. The only parameter is the image tag of the artifact
                        about to be processed.
        :param workers: The same as for the superclass initializer.
        """
        with open(image_tags_file) as f:
            image_tags = list(map(str.strip, f.readlines()))
        super().__init__(image_tags, workers)
        self.copy_dir = copy_dir
        self._command = command

    def pre_run(self):
        copy_tree(self.copy_dir, os.path.join(procutils.HOST_SANDBOX, self.copy_dir))

    def process_artifact(self, image_tag: str):
        return procutils.run_artifact(image_tag, command=self._command(image_tag))



import os
import datetime
import json
import pprint
import sys
import secrets
import requests
import requests.auth

from datetime import date
from typing import Dict
from typing import Generator
from typing import List
from urllib.parse import urljoin
from requests import Response

from ..decorators.classproperty import classproperty
from ..decorators.classproperty import classproperty_support
from .. import log
from bugswarm.common.credentials import COMMON_HOSTNAME

__all__ = ['Endpoint', 'DatabaseAPI']

Endpoint = str


@classproperty_support
class DatabaseAPI(object):
    """
    This class encapsulates programmatic access to the BugSwarm metadata database via the REST API.
    """
    # The base URL must include 'www'. Otherwise, the redirect performed by the backend will cause the requests library
    # to strip authorization headers, which are needed for authentication.
    # See https://github.com/requests/requests/issues/2949 for more information.
    _HOSTNAME = COMMON_HOSTNAME
    if not _HOSTNAME:
        print('[ERROR]: COMMON_HOSTNAME has not been found. Please set the environmental variable and reinstall '
              'common package')
        sys.exit(1)
    _API_VERSION = 'v1'
    _BASE_URL = 'http://{}/{}'.format(_HOSTNAME, _API_VERSION)
    _ARTIFACTS_RESOURCE = 'artifacts'
    _MINED_BUILD_PAIRS_RESOURCE = 'minedBuildPairs'
    _MINED_PROJECTS_RESOURCE = 'minedProjects'
    _EMAIL_SUBSCRIBERS_RESOURCE = 'emailSubscribers'
    _ACCOUNTS_RESOURCE = 'accounts'
    _LOGS_RESOURCE = 'logs'

    def __init__(self, token=None):
        """
        Provide a optional authentication token in order to use the endpoints accessible by the account that is
        associated with the token.
        :param token: An authentication token.
        :raises TypeError: When input token isn't a string or None.
        """
        if token is not None and not isinstance(token, str):
            raise TypeError
        self.token = token
        self.auth = requests.auth.HTTPBasicAuth(username=self.token, password='') if token else None

    ###################################
    # Class properties
    ###################################

    @classproperty
    def base_url(cls) -> Endpoint:
        """
        Exposes the base URL of the API as a read-only class property.
        """
        return cls._BASE_URL

    @classproperty
    def hostname(cls) -> Endpoint:
        """
        Exposes the hostname of the API as a read-only class property.
        """
        return cls._HOSTNAME

    ###################################
    # Artifact REST methods
    ###################################

    def insert_artifact(self, artifact) -> Response:
        return self._insert(DatabaseAPI._artifacts_endpoint(), artifact, 'artifact')

    def patch_artifact(self, image_tag, updates) -> Response:
        if not image_tag:
            raise ValueError
        return self._patch(DatabaseAPI._artifact_image_tag_endpoint(image_tag), updates)

    def find_artifact(self, image_tag: str, error_if_not_found: bool = True) -> Response:
        """
        Get artifact data based on image_tag.
        :param image_tag: The image tag identifying the artifact to find.
        :param error_if_not_found: return err if the image tag not found. default True.
        :return: The response object.
        e.g. find_artifact("Abjad-abjad-289716771")
        """
        log.debug('Trying to find artifact with image_tag {}.'.format(image_tag))
        return self._get(DatabaseAPI._artifact_image_tag_endpoint(image_tag), error_if_not_found)

    def list_artifacts(self) -> List:
        """
        Return a List of artifacts that has at least one reproduce_successes.
        """
        api_filter = '{"reproduce_successes":{"$gt":0}}'
        return self.filter_artifacts(api_filter)

    def filter_artifacts(self, api_filter: str) -> List:
        """
        Return a list of artifacts based on custom filter.
        :param api_filter: a string filter.
        e.g.: api_filter = '{"reproduce_successes":{"$gt":0},"lang":{"$in":["Java","Python"]}}'
        """
        return self._filter(DatabaseAPI._artifacts_endpoint(), api_filter)

    def count_artifacts(self) -> int:
        return self._count(DatabaseAPI._artifacts_endpoint())

    def set_artifact_metric(self, image_tag: str, metric_name: str, metric_value) -> Response:
        """
        Add a metric to an existing artifact.
        The value of the metric can be any valid database type.

        If the metric with name `metric_name` already exists, then its value will be overridden with `metric_value`.
        Otherwise, the metric will be created with value `metric_value`.

        :param image_tag: The image tag identifying the artifact to update.
        :param metric_name: The name of the metric to add or update.
        :param metric_value: The new value of the metric.
        :return: The response object.
        """
        if not isinstance(image_tag, str):
            raise TypeError
        if not image_tag:
            raise ValueError
        if not isinstance(metric_name, str):
            raise TypeError
        if not metric_name:
            raise ValueError
        updates = {'metrics.{}'.format(metric_name): metric_value}
        return self._patch(DatabaseAPI._artifact_image_tag_endpoint(image_tag), updates)

    def set_artifact_failed_config(self, image_tag: str, config) -> Response:
        """
        Add the configuration of the job to an existing artifact's failed job metadata.

        :param image_tag: The image tag identifying the artifact to update.
        :param config: The configuration of the job.
        :return: The response object.
        """
        if not isinstance(image_tag, str):
            raise TypeError
        if not image_tag:
            raise ValueError
        updates = {'failed_job.config': config}
        return self._patch(DatabaseAPI._artifact_image_tag_endpoint(image_tag), updates)

    def set_artifact_passed_config(self, image_tag: str, config) -> Response:
        """
        Add the configuration of the job to an existing artifact's passed job metadata.

        :param image_tag: The image tag identifying the artifact to update.
        :param config: The configuration of the job.
        :return: The response object.
        """
        if not isinstance(image_tag, str):
            raise TypeError
        if not image_tag:
            raise ValueError
        updates = {'passed_job.config': config}
        return self._patch(DatabaseAPI._artifact_image_tag_endpoint(image_tag), updates)

    def set_artifact_classification_exceptions(self, image_tag: str, exceptions) -> Response:
        """
        Add the code exceptions to an existing artifact's classification metadata.

        :param image_tag: The image tag identifying the artifact to update.
        :param exceptions: The standard code exceptions for the artifact.
        :return: The response object.
        """
        if not isinstance(image_tag, str):
            raise TypeError
        if not image_tag:
            raise ValueError
        updates = {'classification.exceptions': exceptions}
        return self._patch(DatabaseAPI._artifact_image_tag_endpoint(image_tag), updates)

    def set_artifact_classification_category(self, image_tag: str, category_type: str, category_confidence) -> Response:
        """
        Add a classification to an existing artifact.

        If the metric with name `category_type` already exists, then its value will be overridden with
        `category_confidence`. Otherwise, the metric will be created with value `category_confidence`.

        :param image_tag: The image tag identifying the artifact to update.
        :param category_type: The name of the category type to update.
        :param category_confidence: The new confidence value of the category.
        :return: The response object.
        """
        if not isinstance(image_tag, str):
            raise TypeError
        if not image_tag:
            raise ValueError
        if not isinstance(category_type, str):
            raise TypeError
        if not category_type:
            raise ValueError
        updates = {'classification.{}'.format(category_type): category_confidence}
        return self._patch(DatabaseAPI._artifact_image_tag_endpoint(image_tag), updates)

    def set_artifact_failed_patch(self, image_tag: str, patch_name: str) -> Response:
        """
        Add the patch for the artifact to an existing artifact's failed job metadata.

        :param image_tag: The image tag identifying the artifact to update.
        :param patch_name: Name of patch that was applied
        :param patch: The patch that was added to the artifact.
        :return: The response object.
        """
        if not isinstance(image_tag, str):
            raise TypeError
        if not image_tag:
            raise ValueError
        today = str(date.today())
        updates = {'failed_job.patches.{}'.format(patch_name): today}
        return self._patch(DatabaseAPI._artifact_image_tag_endpoint(image_tag), updates)

    def set_artifact_passed_patch(self, image_tag: str, patch_name: str) -> Response:
        """
        Add the patch for the artifact to an existing artifact's passed job metadata.

        :param image_tag: The image tag identifying the artifact to update.
        :param patch_name: Name of patch that was applied
        :param patch: The patch that was added to the artifact.
        :return: The response object.
        """
        if not isinstance(image_tag, str):
            raise TypeError
        if not image_tag:
            raise ValueError
        today = str(date.today())
        updates = {'passed_job.patches.{}'.format(patch_name): today}
        return self._patch(DatabaseAPI._artifact_image_tag_endpoint(image_tag), updates)

    def set_artifact_reproduce_successes(self, image_tag: str, successes: int) -> Response:
        """
        Update the reproduce_successes of the job to an existing artifact's failed job metadata.

        :param image_tag: The image tag identifying the artifact to update.
        :param successes: The number of reproduce_successes of the job.
        :return: The response object.
        """
        if not isinstance(image_tag, str):
            raise TypeError
        if not image_tag:
            raise ValueError
        updates = {'reproduce_successes': int(successes)}
        return self._patch(DatabaseAPI._artifact_image_tag_endpoint(image_tag), updates)

    def set_artifact_stability(self, image_tag: str, stability) -> Response:
        """
        Update the stability of the job to an existing artifact's failed job metadata.

        :param image_tag: The image tag identifying the artifact to update.
        :param stability: The reproduce_successes the job.
        :return: The response object.
        """
        if not isinstance(image_tag, str):
            raise TypeError
        if not image_tag:
            raise ValueError
        updates = {'stability': stability}
        return self._patch(DatabaseAPI._artifact_image_tag_endpoint(image_tag), updates)

    def set_artifact_current_status(self, image_tag: str, status: str, date: str) -> Response:
        """
        Update the current status of an artifact.

        :param image_tag: The image tag identifying the artifact to update.
        :param status: Unreproducible/Reproducible/Broken/Flaky.
        :param date: the date (YYYY-MM-DD) of artifacts status changed
        :return: The response object.
        """
        if not isinstance(image_tag, str):
            raise TypeError
        if not image_tag:
            raise ValueError
        if status not in ['Unreproducible', 'Reproducible', 'Broken', 'Flaky']:
            raise ValueError("Incorrect status, should be Unreproducible/Reproducible/Broken/Flaky")
        try:
            datetime.datetime.strptime(date, '%Y-%m-%d')
        except ValueError:
            raise ValueError("Incorrect data format, should be YYYY-MM-DD")
        updates = {'current_status': {'status': status, 'time_stamp': date}}
        return self._patch(DatabaseAPI._artifact_image_tag_endpoint(image_tag), updates)

    def update_artifact_repo_name(self, image_tag: str, new_repo: str) -> Response:
        """
        Update the repository name of an artifact.
        :param image_tag: The image tag identifying the artifact to update.
        :param new_repo: The updated repo name we are setting.
        :return: The response object.
        """
        if not isinstance(image_tag, str):
            raise TypeError
        if not isinstance(new_repo, str):
            raise TypeError
        updates = {'repo': new_repo}
        return self._patch(DatabaseAPI._artifact_image_tag_endpoint(image_tag), updates)

    ###################################
    # Logs REST methods
    ###################################
    def set_build_log(self, job_id: str, build_log_fp: str) -> Response:
        """
        Add the failed or passed build log for the artifact to the logs collection.
        :param job_id: The job id corresponding to the passed or failed build log.
        :param build_log_fp: The path for the build log to be loaded into the database.
        :return: The response object.
        """
        if not isinstance(job_id, str):
            raise TypeError
        if not job_id:
            raise ValueError
        if not isinstance(build_log_fp, str):
            raise TypeError
        if not build_log_fp:
            raise ValueError
        if not os.path.exists(build_log_fp):
            raise IOError
        with open(build_log_fp, 'r') as build_log:
            build_log_text = build_log.read()
        log_entry = {
            'job_id': job_id,
            'build_log': build_log_text,
        }
        return self._insert(DatabaseAPI._logs_endpoint(), log_entry, 'log')

    def get_build_log(self, job_id: str, error_if_not_found: bool = True) -> Response:
        """
        Get artifact failed or passed build log based on job id.
        :param job_id: The job id corresponding to the passed or failed build log.
        :param error_if_not_found: return err if the image tag not found. default True.
        :return: The build_log.
        """
        if not isinstance(job_id, str):
            raise TypeError
        if not job_id:
            raise ValueError
        log_object = self._get(DatabaseAPI._logs_job_id_endpoint(job_id), error_if_not_found).json()
        return log_object['build_log']

    ###################################
    # Mined Build Pair REST methods
    ###################################
    def patch_job_pairs(self, obj_id: str, job_pairs) -> Response:
        updates = {'jobpairs': job_pairs}
        return self._patch(DatabaseAPI._mined_build_pair_object_id_endpoint(obj_id), updates)

    def insert_mined_build_pair(self, mined_build_pair) -> Response:
        return self._insert(DatabaseAPI._mined_build_pairs_endpoint(), mined_build_pair, 'mined build pair')

    def bulk_insert_mined_build_pairs(self, mined_build_pairs) -> bool:
        return all(self._bulk_insert(DatabaseAPI._mined_build_pairs_endpoint(), mined_build_pairs, 'mined build pairs'))

    def find_mined_build_pair(self, object_id: str, error_if_not_found: bool = True) -> Response:
        log.debug('Trying to find mined build pairs for ObjectId {}.'.format(object_id))
        return self._get(DatabaseAPI._mined_build_pair_object_id_endpoint(object_id), error_if_not_found)

    def list_mined_build_pairs(self) -> List:
        return self._list(DatabaseAPI._mined_build_pairs_endpoint())

    def filter_mined_build_pairs(self, api_filter: str) -> List:
        return self._filter(DatabaseAPI._mined_build_pairs_endpoint(), api_filter)

    def count_mined_build_pairs(self) -> int:
        return self._count(DatabaseAPI._mined_build_pairs_endpoint())

    def filter_mined_build_pairs_for_repo(self, repo: str) -> List:
        """
        Returns a list of build pairs mined from `repo`. Returns an empty list if no build pairs for `repo` are found or
        if `repo` has not yet been mined.
        """
        return self._filter(DatabaseAPI._mined_build_pairs_endpoint(), '{{"repo": "{}"}}'.format(repo))

    def remove_mined_build_pairs_for_repo(self, repo: str) -> bool:
        """
        Non-atomically remove the existing mined build pairs for `repo`.

        If your goal is to replace mined build pairs for `repo`, you should use `replace_mined_build_pairs_for_repo`.
        """
        for bp in self.filter_mined_build_pairs_for_repo(repo):
            if not self._delete(DatabaseAPI._mined_build_pair_object_id_endpoint(bp['_id'])):
                log.error('Could not remove an existing mined build pair for {}.'.format(repo))
                return False
        return True

    def replace_mined_build_pairs_for_repo(self, repo: str, new_build_pairs: List[Dict]) -> bool:
        """
        Non-atomically remove the existing mined build pairs for `repo` and then non-atomically insert the newly mined
        build pairs for `repo` in `new_build_pairs`.
        """
        if not self.remove_mined_build_pairs_for_repo(repo):
            return False
        # Take advantage of bulk insertions.
        if not self.bulk_insert_mined_build_pairs(new_build_pairs):
            log.error('While replacing mined build pairs, a bulk insertion failed.')
            return False
        return True

    def update_mined_build_pairs_repo_name(self, obj_id: str, new_repo: str) -> Response:
        """
        Update the repository name of an artifact in the minedBuildPairs db.
        :param obj_id: The unique id of an object.
        :param new_repo: The updated repo name we are setting.
        :return: The response object.
        """
        if not isinstance(new_repo, str):
            raise TypeError
        updates = {'repo': new_repo}
        return self._patch(DatabaseAPI._mined_build_pair_object_id_endpoint(obj_id), updates)

    ###################################
    # Mined Project REST methods
    ###################################

    def insert_mined_project(self, mined_project) -> Response:
        return self._insert(DatabaseAPI._mined_projects_endpoint(), mined_project, 'mined project')

    def find_mined_project(self, repo: str, ci_service: str, error_if_not_found: bool = True) -> Response:
        log.debug('Trying to find mined project with repo {} and CI service {}.'.format(repo, ci_service))
        return self._get(DatabaseAPI._mined_project_repo_endpoint(repo, ci_service), error_if_not_found)

    def list_mined_projects(self) -> List:
        return self._list(DatabaseAPI._mined_projects_endpoint())

    def filter_mined_projects(self, api_filter: str) -> List:
        return self._filter(DatabaseAPI._mined_projects_endpoint(), api_filter)

    # TODO should this count repos, or repo-CI pairs?
    def count_mined_projects(self) -> int:
        return self._count(DatabaseAPI._mined_projects_endpoint())

    def upsert_mined_project(self, mined_project) -> Response:
        """
        Upsert a mined project. Can be used for initial mining or re-mining of a project.
        """
        repo = mined_project.get('repo')
        ci_service = mined_project.get('ci_service')
        assert repo
        assert ci_service
        return self._upsert(DatabaseAPI._mined_project_repo_endpoint(repo, ci_service), mined_project, 'mined project')

    def set_latest_build_info_metric(self, repo: str, ci_service: str, build_number: int, build_id: int):
        """
        Set the build information regarding the build number and build id of the latest build we've mined.

        :param repo: The repository slug for identifying the mined project to update.
        :param ci_service: The CI service for identifying the project to update.
        :param build_number: The latest build number associated to the last build we've mined
        :param build_id: The latest build id associated to the last build we've mined
        :return: The response object.
        """
        if not isinstance(repo, str):
            raise TypeError
        if not repo:
            raise ValueError
        if not isinstance(build_number, int):
            raise TypeError
        if not build_number:
            raise ValueError
        if not isinstance(build_id, int):
            raise TypeError
        if not build_id:
            raise ValueError
        updates = {
            'last_build_mined.build_number': build_number,
            'last_build_mined.build_id': build_id
        }
        return self._patch(DatabaseAPI._mined_project_repo_endpoint(repo, ci_service), updates)

    def set_mined_project_progression_metric(self, repo: str, ci_service: str,
                                             metric_name: str, metric_value) -> Response:
        """
        Add a mining progression metric to an existing mined project.
        The value of the metric can be any valid database type.

        If the metric with name `metric_name` already exists, then its value will be overridden with `metric_value`.
        Otherwise, the metric will be created with value `metric_value`.

        :param repo: The repository slug identifying the mined project to update.
        :param metric_name: The name of the metric to add or update.
        :param metric_value: The new value of the metric.
        :return: The response object.
        """
        if not isinstance(repo, str):
            raise TypeError
        if not repo:
            raise ValueError
        if not isinstance(metric_name, str):
            raise TypeError
        if not metric_name:
            raise ValueError
        updates = {'progression_metrics.{}'.format(metric_name): metric_value}
        return self._patch(DatabaseAPI._mined_project_repo_endpoint(repo, ci_service), updates)

    def update_mined_project_repo_name(self, repo: str, ci_service: str, new_repo: str):
        if not isinstance(repo, str):
            raise TypeError
        if not isinstance(repo, str):
            raise TypeError
        updates = {'repo': new_repo}
        return self._patch(DatabaseAPI._mined_project_repo_endpoint(repo, ci_service), updates)

    def soft_delete_mined_project(self, repo: str, ci_service: str) -> Response:
        if not isinstance(repo, str):
            raise TypeError
        if not repo:
            raise ValueError
        return self._delete(DatabaseAPI._mined_project_repo_endpoint(repo, ci_service))

    ###################################
    # Email Subscriber REST methods
    ###################################

    def insert_email_subscriber(self, email_subscriber) -> Response:
        return self._insert(DatabaseAPI._email_subscribers_endpoint(), email_subscriber, 'email subscriber')

    def find_email_subscriber(self, email: str, error_if_not_found: bool = True) -> Response:
        log.debug('Trying to find email subscriber with email {}.'.format(email))
        return self._get(DatabaseAPI._email_subscriber_email_endpoint(email), error_if_not_found)

    def list_email_subscribers(self) -> List:
        return self._list(DatabaseAPI._email_subscribers_endpoint())

    def filter_email_subscribers(self, api_filter: str) -> List:
        return self._filter(DatabaseAPI._email_subscribers_endpoint(), api_filter)

    def count_email_subscribers(self) -> int:
        return self._count(DatabaseAPI._email_subscribers_endpoint())

    def confirm_email_subscriber(self, email: str) -> Response:
        # Set confirmed to True and clear the confirmation token.
        updates = {'confirmed': True, 'confirm_token': ''}
        return self._patch(DatabaseAPI._email_subscriber_email_endpoint(email), updates)

    def unsubscribe_email_subscriber(self, email: str) -> Response:
        return self._delete(DatabaseAPI._email_subscriber_email_endpoint(email))

    ###################################
    # Account REST methods
    ###################################

    def create_account(self, email: str, role: list = None) -> Response:
        """
       Create an account with given email and role. Return token in Response if succeed.
       :param email: email address.
       :param role: default to ['user'].
       :return: The response object.
       """
        token = secrets.token_urlsafe()
        while self.filter_account_for_token(token):
            token = secrets.token_urlsafe()
        if role is None:
            role = ['user']
        account = {'email': email, 'roles': role, 'token': token, 'password': ''}
        response = self._insert(DatabaseAPI._accounts_endpoint(), account, 'account')
        if response.ok:
            response._content = json.dumps(account).encode('utf-8')
        return response

    def find_account(self, email: str, error_if_not_found: bool = True) -> Response:
        log.debug('Trying to find account with email {}.'.format(email))
        return self._get(DatabaseAPI._account_email_endpoint(email), error_if_not_found)

    def list_accounts(self) -> List:
        return self._list(DatabaseAPI._accounts_endpoint())

    def filter_account_for_token(self, token: str) -> List:
        return self._filter(DatabaseAPI._accounts_endpoint(), '{{"token":"{}"}}'.format(token))

    def filter_accounts(self, api_filter: str) -> List:
        return self._filter(DatabaseAPI._accounts_endpoint(), api_filter)

    def count_accounts(self) -> int:
        return self._count(DatabaseAPI._accounts_endpoint())

    ###################################
    # Convenience REST methods
    ###################################

    def _get(self, endpoint: Endpoint, error_if_not_found: bool = True) -> Response:
        if not isinstance(endpoint, Endpoint):
            raise TypeError
        if not endpoint:
            raise ValueError
        resp = requests.get(endpoint, auth=self.auth)
        # Do not print an error message if the entity was not expected to be found and we got a 404 status code.
        not_found = resp.status_code == 404
        rate_limited = resp.status_code == 429
        if not_found and not error_if_not_found:
            return resp
        elif rate_limited:
            log.error('You are being rate limited. You can request a API token to remove this limit.')
            log.error('http://www.bugswarm.org/contact/')
        elif not resp.ok:
            log.error(resp.url)
            log.error(resp.content)
        return resp

    def _post(self, endpoint: Endpoint, data) -> Response:
        if not isinstance(endpoint, Endpoint):
            raise TypeError
        if not endpoint:
            raise ValueError
        headers = {'Content-Type': 'application/json'}
        resp = requests.post(endpoint, json.dumps(data), headers=headers, auth=self.auth)
        if not resp.ok:
            log.error(resp.url)
            log.error(resp.content)
        return resp

    def _patch(self, endpoint: Endpoint, updates) -> Response:
        if not isinstance(endpoint, Endpoint):
            raise TypeError
        if not endpoint:
            raise ValueError
        # First get the entity's etag.
        etag = self._get(endpoint).json()['_etag']
        # Now patch the entity.
        headers = {
            'Content-Type': 'application/json',
            'If-Match': etag,
        }
        resp = requests.patch(endpoint, json.dumps(updates), headers=headers, auth=self.auth)
        if not resp.ok:
            log.error(resp.url)
            log.error(resp.content)
        return resp

    def _put(self, endpoint: Endpoint, data, etag: str = None) -> Response:
        if not isinstance(endpoint, Endpoint):
            raise TypeError
        if not endpoint:
            raise ValueError
        headers = {'Content-Type': 'application/json'}
        if etag:
            headers['If-Match'] = etag
        # Now replace the entity.
        resp = requests.put(endpoint, json.dumps(data), headers=headers, auth=self.auth)
        if not resp.ok:
            log.error(resp.url)
            log.error(resp.content)
        return resp

    def _delete(self, endpoint: Endpoint) -> Response:
        if not isinstance(endpoint, Endpoint):
            raise TypeError
        if not endpoint:
            raise ValueError
        # First get the entity's etag.
        etag = self._get(endpoint).json()['_etag']
        headers = {'If-Match': etag}
        # Now delete the entity.
        resp = requests.delete(endpoint, headers=headers, auth=self.auth)
        if not resp.ok:
            log.error(resp.url)
            log.error(resp.content)
        return resp

    ###################################
    # Convenience methods
    ###################################

    def _insert(self, endpoint: Endpoint, entity, singular_entity_name: str = 'entity') -> Response:
        if entity is None:
            raise TypeError
        if not isinstance(endpoint, Endpoint):
            raise TypeError
        if not endpoint:
            raise ValueError
        if not isinstance(singular_entity_name, str):
            raise TypeError
        if not singular_entity_name:
            raise ValueError
        log.debug('Trying to insert {}.'.format(singular_entity_name))
        resp = self._post(endpoint, entity)
        if resp.status_code == 422:
            log.error('The', singular_entity_name, 'was not inserted because it failed validation.')
            log.error(pprint.pformat(entity))
            log.error(resp.content)
        return resp

    def _bulk_insert(self,
                     endpoint: Endpoint,
                     entities: List,
                     plural_entity_name: str = 'entities') -> Generator[Response, None, None]:
        if not isinstance(entities, list):
            raise TypeError('The {} to bulk insert must be provided as a list.'.format(plural_entity_name))
        if not all(e for e in entities):
            raise ValueError('All {} must be non-None.'.format(plural_entity_name))
        if not isinstance(endpoint, Endpoint):
            raise TypeError
        if not endpoint:
            raise ValueError
        if not isinstance(plural_entity_name, str):
            raise TypeError
        if not plural_entity_name:
            raise ValueError
        log.debug('Trying to bulk insert {} {}.'.format(len(entities), plural_entity_name))
        # Insert the entities in chunks to avoid a 413 Request Entity Too Large error.
        for chunk in DatabaseAPI._chunks(entities, 100):
            list_for_insertion = []
            list_of_build_ids = []
            for data in chunk:
                failed_build_id = data['failed_build']['build_id']
                list_of_build_ids.append(failed_build_id)

            api_filter = '{"failed_build.build_id":{"$in":' + '{}'.format(list_of_build_ids) + '}}'
            list_of_existing_build_pairs = self.filter_mined_build_pairs(api_filter)
            for data in chunk:
                bp_exists = False
                for bp in list_of_existing_build_pairs:
                    if data['failed_build']['build_id'] == bp['failed_build']['build_id']:
                        log.info('Failed Build ID: {} already exists in the database. Skipping insertion.'
                                 .format(data['failed_build']['build_id']))
                        bp_exists = True
                        break
                if bp_exists:
                    continue
                list_for_insertion.append(data)
            resp = self._post(endpoint, list_for_insertion)
            if resp.status_code == 422:
                log.error('The', plural_entity_name, 'were not inserted because they failed validation.')
                log.error(pprint.pformat(chunk))
                log.error(resp.content)
            elif resp.status_code == 400:
                log.error('Buildpairs were not inserted because the list is either empty, or all pairs already exist.')
            yield resp

    def _upsert(self, endpoint: Endpoint, entity, singular_entity_name: str = 'entity') -> Response:
        if entity is None:
            raise TypeError
        if not isinstance(endpoint, Endpoint):
            raise TypeError
        if not endpoint:
            raise ValueError
        if not isinstance(singular_entity_name, str):
            raise TypeError
        if not singular_entity_name:
            raise ValueError
        log.debug('Trying to upsert {}.'.format(singular_entity_name))
        # First, check if the entity exists. If it does, then pass its etag.
        get_resp = self._get(endpoint, error_if_not_found=False)
        etag = get_resp.json()['_etag'] if get_resp.ok else None
        resp = self._put(endpoint, entity, etag)
        if resp.status_code == 422:
            log.error('The', singular_entity_name, 'was not upserted because it failed validation.')
            log.error(pprint.pformat(entity))
            log.error(resp.content)
        return resp

    def _iter_pages(self, start_link: str) -> List:
        """
        Returns a list of all the results by following the next link chain starting with `start_link`.
        """
        if not isinstance(start_link, str):
            raise TypeError
        if not start_link:
            raise ValueError
        results = []
        next_link = start_link
        while next_link:
            next_resp = self._get(next_link)
            next_resp.raise_for_status()
            next_json = next_resp.json()
            try:
                if not next_json['_items']:
                    break
                results += next_json['_items']
            except KeyError:
                break
            try:
                next_link = urljoin(next_link, next_json['_links']['next']['href'])
            except KeyError:
                break
        return results

    def _list(self, endpoint: Endpoint) -> List:
        """
        Returns all results from the current page to the last page, inclusive.
        """
        if not isinstance(endpoint, Endpoint):
            raise TypeError
        if not endpoint:
            raise ValueError
        return self._iter_pages(endpoint)

    def _filter(self, endpoint: Endpoint, api_filter: str) -> List:
        if not isinstance(endpoint, Endpoint):
            raise TypeError
        if not endpoint:
            raise ValueError
        if not isinstance(api_filter, str):
            raise TypeError
        if not api_filter:
            raise ValueError
        # Append the filter as a url parameter.
        url = '{}?where={}'.format(endpoint, api_filter)
        return self._iter_pages(url)

    def _count(self, endpoint: Endpoint) -> int:
        if not isinstance(endpoint, Endpoint):
            raise TypeError
        if not endpoint:
            raise ValueError
        resp = self._get(endpoint)
        result = resp.json()
        try:
            total = result['_meta']['total']
            assert isinstance(total, int)
            return total
        except KeyError:
            return -1

    @staticmethod
    def _chunks(l: List, n: int) -> Generator[List, None, None]:
        """
        Yields successive `n`-sized chunks from `l`. `n` is clamped to the length of `l`.
        """
        if not isinstance(l, list):
            raise TypeError
        if not isinstance(n, int):
            raise TypeError
        if n == 0:
            raise ValueError('Chunk size must be greater than zero.')
        if not l:
            # Yield nothing if the list is empty.
            return
        # Even though clamping is implied when slicing, we explicitly clamp n to the length of the list.
        n = min(n, len(l))
        for i in range(0, len(l), n):
            yield l[i:i + n]

    @staticmethod
    def _endpoint(resource: str) -> Endpoint:
        if not isinstance(resource, str):
            raise TypeError
        if not resource:
            raise ValueError
        return '/'.join([DatabaseAPI.base_url, resource])

    @staticmethod
    def _artifacts_endpoint() -> Endpoint:
        return DatabaseAPI._endpoint(DatabaseAPI._ARTIFACTS_RESOURCE)

    @staticmethod
    def _artifact_image_tag_endpoint(image_tag: str) -> Endpoint:
        if not isinstance(image_tag, str):
            raise TypeError
        if not image_tag:
            raise ValueError
        return '/'.join([DatabaseAPI._artifacts_endpoint(), image_tag])

    @staticmethod
    def _mined_build_pairs_endpoint() -> Endpoint:
        return DatabaseAPI._endpoint(DatabaseAPI._MINED_BUILD_PAIRS_RESOURCE)

    @staticmethod
    def _mined_build_pair_object_id_endpoint(object_id: str) -> Endpoint:
        if not isinstance(object_id, str):
            raise TypeError
        if not object_id:
            raise ValueError
        return '/'.join([DatabaseAPI._mined_build_pairs_endpoint(), object_id])

    @staticmethod
    def _mined_projects_endpoint() -> Endpoint:
        return DatabaseAPI._endpoint(DatabaseAPI._MINED_PROJECTS_RESOURCE)

    @staticmethod
    def _mined_project_repo_endpoint(repo: str, ci_service) -> Endpoint:
        if not isinstance(repo, str):
            raise TypeError
        if not repo:
            raise ValueError
        if ci_service not in ['travis', 'github']:
            raise ValueError
        return '/'.join([DatabaseAPI._mined_projects_endpoint(), ci_service, repo])

    @staticmethod
    def _email_subscribers_endpoint() -> Endpoint:
        return DatabaseAPI._endpoint(DatabaseAPI._EMAIL_SUBSCRIBERS_RESOURCE)

    @staticmethod
    def _email_subscriber_email_endpoint(email: str) -> Endpoint:
        if not isinstance(email, str):
            raise TypeError
        if not email:
            raise ValueError
        return '/'.join([DatabaseAPI._email_subscribers_endpoint(), email])

    @staticmethod
    def _accounts_endpoint() -> Endpoint:
        return DatabaseAPI._endpoint(DatabaseAPI._ACCOUNTS_RESOURCE)

    @staticmethod
    def _account_email_endpoint(email: str) -> Endpoint:
        if not isinstance(email, str):
            raise TypeError
        if not email:
            raise ValueError
        return '/'.join([DatabaseAPI._accounts_endpoint(), email])

    @staticmethod
    def _logs_endpoint() -> Endpoint:
        return DatabaseAPI._endpoint(DatabaseAPI._LOGS_RESOURCE)

    @staticmethod
    def _logs_job_id_endpoint(job_id: str) -> Endpoint:
        if not isinstance(job_id, str):
            raise TypeError
        if not job_id:
            raise ValueError
        return '/'.join([DatabaseAPI._logs_endpoint(), job_id])



import os
import time
import urllib.request

from concurrent.futures import as_completed
from concurrent.futures import ThreadPoolExecutor
from typing import List
from typing import Union
from urllib.error import URLError

from bugswarm.common import log

_DEFAULT_RETRIES = 3


def download_log(job_id: Union[str, int],
                 destination: str,
                 overwrite: bool = True,
                 retries: int = _DEFAULT_RETRIES) -> bool:
    """
    Downloads a Travis job log and stores it at destination.

    :param job_id: Travis job ID, as a string or integer, identifying a job whose log to download.
    :param destination: Path where the log should be stored.
    :param overwrite: Whether the file at `destination`, if it already exists, should be replaced.
    :param retries: The number of times to retry the log download if it fails. Defaults to 3, in which case the network
                    will be accessed up to 4 (3 + 1) times.
    :param overwrite: Whether to overwrite a file at `destination` if one already exists. If `overwrite` is False and a
                      file already exists at `destination`, FileExistsError is raised. Defaults to True.
    :raises ValueError:
    :raises FileExistsError: When a file at `destination` already exists and `overwrite` is False.
    :return: True if the download succeeded.
    """
    if not job_id:
        raise ValueError
    if not destination:
        raise ValueError
    if os.path.isfile(destination) and not overwrite:
        log.error('The log for job', job_id, 'already exists locally.')
        raise FileExistsError

    job_id = str(job_id)
    travis_log_link = 'https://api.travis-ci.org/jobs/{}/log.txt'.format(job_id)
    content = _get_log_from_url(travis_log_link, retries)

    if not content:
        travis_log_link = 'https://api.travis-ci.com/v3/job/{}/log.txt'.format(job_id)
        content = _get_log_from_url(travis_log_link, retries)
        # If this endpoint fails, the log is not on either endpoint and does not exist
        if not content:
            return False

    with open(destination, 'wb') as f:
        f.write(content)
    return True


def download_logs(job_ids: List[Union[str, int]],
                  destinations: List[str],
                  overwrite: bool = True,
                  num_workers: int = 5,
                  retries: int = _DEFAULT_RETRIES) -> bool:
    """
    Downloads one or more Travis job logs in parallel and stores them at the given destinations.
    This function calls `download_log` and raises the first exception it catches from that function, if any.

    If you only need to download a single Travis job log, use the `download_log` function.

    :param job_ids: A list of Travis job IDs, as strings or integers, identifying jobs whose logs to download.
    :param destinations: A list of paths where the logs should be stored. The path at index `i` corresponds to the log
                         downloaded for the job ID at index `i` in `job_ids`. Thus, `job_ids` and `destinations` must be
                         the same length.
    :param overwrite: Same as the argument for `download_log`.
    :param num_workers: Number of workers to download logs. Defaults to the maximum of 5.
    :param retries: Same as the argument for `download_log`.
    :raises ValueError:
    :raises FileExistsError: When a file already exists at the given destination and `overwrite` is False.
    :return: True if all downloads succeeded.
    """
    if not job_ids:
        raise ValueError
    if not destinations:
        raise ValueError
    if not len(job_ids) == len(destinations):
        log.error('The job_ids and destinations arguments must be of equal length.')
        raise ValueError

    num_workers = min(num_workers, len(job_ids))
    with ThreadPoolExecutor(max_workers=num_workers) as executor:
        future_to_job_id = {executor.submit(download_log, job_id, dst, overwrite, retries): job_id
                            for job_id, dst in zip(job_ids, destinations)}

    succeeded = 0
    for future in as_completed(future_to_job_id):
        try:
            # The result will be True if the download succeeded. Otherwise, future.result() will raise an exception or
            # return False.
            ok = future.result()
        except Exception:
            raise
        else:
            if ok:
                succeeded += 1

    return succeeded == len(job_ids)


def _get_log_from_url(log_url: str, max_retries: int, retry_count: int = 0):
    sleep_duration = 3  # Seconds.
    try:
        with urllib.request.urlopen(log_url) as url:
            result = url.read()
            log.info('Downloaded log from {}.'.format(log_url))
            return result
    except URLError as e:
        log.error('Could not download log from {}.'.format(log_url, e.reason))
        return None
    except ConnectionResetError:
        if retry_count == max_retries:
            log.warning('Could not download log from', log_url, 'after retrying', max_retries, 'times.')
            return None
        log.warning('The server reset the connection. Retrying after', sleep_duration, 'seconds.')
        time.sleep(sleep_duration)
        _get_log_from_url(log_url, max_retries, retry_count + 1)



import json
import os

from typing import Optional


class DefaultJSONEncoder(json.JSONEncoder):
    """
    Subclass of `json.JSONEncoder` that can serialize user-defined types.
    """
    def default(self, o):
        return o.__dict__


def read_json(filepath: str):
    """
    Decode JSON data from a file.

    :param filepath: Path to the file from which to read JSON data.
    :raises FileNotFoundError: If no file exists at `filepath`.
    :return: The decoded JSON object.
    """
    if not filepath:
        raise ValueError
    if not os.path.isfile(filepath):
        raise FileNotFoundError
    with open(filepath) as f:
        return json.load(f)


def write_json(filepath: str, data, overwrite: bool = True, encoder: Optional[json.JSONEncoder] = None):
    """
    Write JSON data to a file.

    :param filepath: Path to the file to which to write JSON data.
    :param data: The JSON object to encode and write to the file.
    :param overwrite: Whether to overwrite a file at `filepath` if one already exists. If `overwrite` is False and a
                      file already exists at `filepath`, FileExistsError is raised. Defaults to True.
    :param encoder: The encoder object to use for the JSON encoding. Should be a subclass of json.JSONEncoder.
                    DefaultJSONEncoder is used by default.
    :raises FileExistsError: If a file at `filepath` already exists and `overwrite` is False.
    :return: True if the write to `filepath` succeeded.
    """
    if not filepath:
        raise ValueError
    if os.path.isfile(filepath) and not overwrite:
        raise FileExistsError
    if data is None:
        raise ValueError
    with open(filepath, 'w') as f:
        json.dump(data, f, cls=encoder, indent=2)



"""Setup that only installs bugswarm.common and its dependencies. Used
to generate the bugswarm-common package on PyPI."""

from datetime import datetime

from setuptools import setup, find_packages


today = datetime.utcnow()
version = f'{today.year}.{today.month:02}.{today.day:02}'

setup(
    name='bugswarm-common',
    version=version,
    url='https://github.com/BugSwarm/bugswarm',
    author='BugSwarm',
    author_email='dev.bugswarm@gmail.com',

    description='Library of modules used throughout the BugSwarm toolset',
    long_description='Library of modules used throughout the BugSwarm toolset',
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: BSD License',
    ],
    zip_safe=False,
    packages=find_packages(include=['bugswarm.common*']),
    install_requires=[
        'requests>=2.20.0',
        'CacheControl==0.12.3',
        'requests-cache==0.4.13',
    ],
)



import time
from collections import deque

from typing import List
from typing import Optional
from typing import Tuple
from urllib.parse import urlparse

import cachecontrol
import requests
import copy

from bugswarm.common import log


class GitHubWrapper(object):
    """
    Wrapper around the GitHub API. This wraper is not meant to implement convenience methods to access GitHub API
    endpoints. Instead, its goals are to facilitate:
    - automatically retrying when a GitHub request fails
    - automatically choosing another available GitHub token when the current token's quota is exceeded
    - automatically waiting until any token's quota is reset when all tokens exceeded their quota.
    """
    def __init__(self, tokens: List[str]):
        """
        :param tokens: A list of GitHub tokens.
        """
        if not isinstance(tokens, list):
            raise TypeError('Tokens must be a list.')
        if not all(isinstance(t, str) for t in tokens):
            raise ValueError('All GitHub tokens must be given as strings.')

        self._tokens = deque(tokens)
        self._session = cachecontrol.CacheControl(requests.Session())
        # Start with the first token. We lazily switch tokens as each hits its quota limit.
        self._session.headers['Authorization'] = 'token %s' % self._tokens[0]

    def get(self, url: str):
        """
        Request a URL from the GitHub API.
        Handles retrying, waiting for quota to reset, and token switching.

        :param url: The GitHub API URL to request.
        :return: A 2-tuple of the resulting response and the JSON representation of the response body. If there was a
                 problem, the returned tuple is (None, None).
        """
        if not isinstance(url, str):
            raise TypeError('The provided URL must be a string.')
        if urlparse(url).netloc != 'api.github.com':
            raise ValueError('The provided URL is not for the GitHub API.')

        retry_back_off = 5  # Seconds.
        retry_count = 0
        while True:
            response = None
            try:
                response = self._session.get(url)
                response.raise_for_status()
                if not response.text:
                    return None, None
                return response, response.json()
            except Exception as e:
                # If the exception is a connection error, the server may have dropped the connection.
                # In this case, we should try resetting the session.
                if e is requests.ConnectionError:
                    log.info('Recreating session.')
                    self._create_session()

                if response.status_code == 404:
                    return None, None
                elif response.status_code == 451:  # Repository access blocked.
                    log.error('Repository access blocked:', url)
                    return None, None
                elif response.status_code == 401:  # Not authorized.
                    log.error('Invalid GitHub API token: ', self._session.headers['Authorization'])
                    return None, None
                elif response.status_code == 422:
                    return None, None
                else:
                    log.error('Request for url failed:', url)
                    log.error('Exception:', e)

                # If the status code is 403 (Forbidden), then we may have exceeded our GitHub API quota.
                # In this case, we should verify that the quota was exceeded and, if so, wait until the quota is reset.
                if response is not None and response.status_code == 403:
                    result = response.json()
                    # Check whether GitHub's abuse detection mechanism was triggered.
                    if 'message' in result:
                        if 'abuse detection mechanism' in result['message']:
                            log.warning('Triggered the GitHub abuse detection mechanism. Sleeping for 1 minute.')
                            time.sleep(60)
                        if 'Not Found' == result['message']:
                            return None, None

                    quota_exceeded, sleep_duration = self._exceeded_api_quota()
                    if quota_exceeded:
                        # Pick another token.
                        self._create_session()
                time.sleep(retry_back_off)
                retry_count += 1

    def get_all_pages(self, url: str):
        """
        Request a URL from the GitHub API that requires pagination.
        Handles retrying, waiting for quota to reset, and token switching.

        :param url: The GitHub API URL to request. Should require pagination.
        :return: The concatenated results of all pages of results for `url`.
        """
        if not isinstance(url, str):
            raise TypeError('The provided URL must be a string.')

        all_results = []
        while True:
            response, result = self.get(url)
            if None in (response, result):
                break
            all_results = sum([result], all_results)
            next_reference = response.links.get('next')
            if next_reference is None:
                break
            url = next_reference.get('url')
        return all_results

    def _exceeded_api_quota(self) -> Tuple[bool, Optional[int]]:
        """
        :return: A 2-tuple. (True, number of seconds until the quota resets) if the API quota has been exceeded.
                 (False, None) otherwise.
        :raises Exception: When an exception is raised by the request.
        """
        quota_url = 'https://api.github.com/rate_limit'
        log.info('Checking GitHub API quota.')
        response = self._session.get(quota_url)
        try:
            response.raise_for_status()
            result = response.json()
            if 'resources' in result:
                remaining = result['resources']['core']['remaining']
                if remaining <= 0:
                    reset_at = result['resources']['core']['reset']  # Time when the quota resets, in UTC epoch seconds
                    log.warning('GitHub API quota exceeded and will reset at UTC {}.'.format(reset_at))
                    now = int(time.time())
                    sleep_duration = (reset_at - now) + 10  # Add a few seconds to be sure that we sleep long enough.
                    return True, sleep_duration
        except Exception as e:
            log.error('Exception while checking API quota:', e)
            raise
        return False, None

    def _create_session(self):
        """
        When the quota is exceeded for a token, the program will switch to another tokens and attempt to continue.
        If the quota is exceeded for all tokens, the program will wait for the token with the lowest wait time.
        """
        min_wait_time = 9999
        chosen_token = None
        updated_token = copy.deepcopy(self._tokens)
        for t in self._tokens:
            self._session = cachecontrol.CacheControl(requests.Session())
            self._session.headers['Authorization'] = 'token %s' % t
            has_wait, wait_time = self._exceeded_api_quota()
            if not has_wait:
                chosen_token = t
                min_wait_time = 0
                # if a token is chosen, move it to the end
                updated_token.append(t)
                del updated_token[updated_token.index(t)]
                break
            if wait_time < min_wait_time:
                min_wait_time = wait_time
                chosen_token = t
                # if a token is chosen, move it to the end
                updated_token.append(t)
                del updated_token[updated_token.index(t)]
        self._tokens = updated_token
        if not chosen_token:
            raise RuntimeError('Unexpected state: No GitHub token chosen in github.py.')
        log.debug('Chose token {}.'.format(chosen_token))
        if min_wait_time:
            # Sleep until the quota is reset. See https://developer.github.com/v3/#rate-limiting for more information.
            log.warning('Sleeping until the GitHub API quota is reset in', min_wait_time / 60, 'minutes.')
            time.sleep(min_wait_time)
        self._session = cachecontrol.CacheControl(requests.Session())
        self._session.headers['Authorization'] = 'token %s' % chosen_token
